"""Utility functions for LSST JupyterLab notebook environment
"""
import bokeh
import os
import urllib


def show_with_bokeh_server(obj):
    """Method to wrap bokeh with proxy URL
    """
    def jupyter_proxy_url(port):
        """Construct proxy URL from environment
        """
        # If port is None we're asking about the URL
        # for the origin header.
        if port is None:
            return '*'

        base_url = os.environ['EXTERNAL_URL']
        service_url_path = os.environ['JUPYTERHUB_SERVICE_PREFIX']
        proxy_url_path = 'proxy/%d' % port

        user_url = urllib.parse.urljoin(base_url, service_url_path)
        full_url = urllib.parse.urljoin(user_url, proxy_url_path)
        return full_url

    bokeh.io.show(obj, notebook_url=jupyter_proxy_url)



"""
Collection of notebook utilities
"""
from .utils import show_with_bokeh_server
from .forwarder import Forwarder
from .clusterproxy import ClusterProxy
__all__ = [show_with_bokeh_server, Forwarder, ClusterProxy]



import functools
import logging
import maproxy.proxyserver
import maproxy.iomanager
import tornado.netutil
import threading
import time


class Forwarder(maproxy.proxyserver.ProxyServer):
    """This creates a TCP proxy server running on a randomly-assigned
    dynamic local port.  Pass it the target host and port to construct it.
    """
    _ioloop = None
    _thread = None
    _logger = None
    _bind_addresses = []

    def __init__(self, *args, **kwargs):
        self._logger = logging.getLogger(__name__)
        sockets = tornado.netutil.bind_sockets(0, '')
        ioloop = None
        if "ioloop" in kwargs:
            ioloop = kwargs["ioloop"]
            ioloop.make_current()
            del kwargs["ioloop"]
        super().__init__(*args, **kwargs)
        self.add_sockets(sockets)
        self.bind_addresses = [x.getsockname()[:2] for x in sockets]
        self._ioloop = tornado.ioloop.IOLoop.current()

    def get_ports(self):
        """Returns a list of the ports the Forwarder is listening to."""
        return list(set(x[1] for x in self.bind_addresses))

    def get_port(self):
        """Returns the first port the Forwarder is listening to, or None."""
        rval = self.get_ports()
        if rval and len(rval) > 0:
            return rval[0]
        return None



import logging
import os
from .forwarder import Forwarder
from dask.distributed import Client


def _remove_prefix(s, prefix):
    return s[len(prefix):] if s.startswith(prefix) else s


def _get_hostname():
    return os.environ.get('HOSTNAME') or 'localhost'


def _get_proxy_url(port):
    urlprefix = os.environ.get("EXTERNAL_URL") or "http://localhost:8888"
    user = os.environ.get("USER") or "jovyan"
    url = "/".join([urlprefix, "user", user, "proxy", str(port)])
    return url


class ClusterProxy(object):
    """Provides a proxy service to map a local port to a worker node's
    dashboard, which is on port 8989.  This allows us to proxy to the
    worker even though the k8s network is not accessible externally.

    It must be created with an instance of a dask.distributed.Client
    as its argument.
    """
    client = None
    cluster = None
    ioloop = None
    workers = {}

    def __init__(self, client):
        if not isinstance(client, Client):
            estr = "'client' argument must be dask.distributed.Client!"
            raise RuntimeError(estr)
        self.client = client
        self.logger = logging.getLogger(__name__)
        self.cluster = client.cluster
        self.ioloop = client.io_loop
        self.scheduler = _get_proxy_url(8787) + "/status"
        self.refresh_workers()

    def refresh_workers(self):
        """Rebuild current worker map from actual state.
        """
        coremap = self.client.ncores()
        current_workerlist = list(coremap.keys())
        current_workerlist = [_remove_prefix(x, "tcp://")
                              for x in current_workerlist]
        removed_workers = []
        for worker_id in self.workers:
            if worker_id not in current_workerlist:
                # Get rid of removed workers
                self._remove_worker_proxy(worker_id)
                removed_workers.append(worker_id)
        for worker_id in removed_workers:
            del self.workers[worker_id]
        for worker_id in current_workerlist:
            if worker_id not in self.workers:
                self.workers[worker_id] = self._create_worker_proxy(worker_id)
            # Otherwise it hasn't changed.

    def _remove_worker_proxy(self, worker_id):
        worker = self.workers.get(worker_id)
        if not worker:
            return
        forwarder = worker["forwarder"]
        forwarder.stop()

    def _create_worker_proxy(self, worker_id):
        host = worker_id.split(":")[0]
        port = 8989
        proxy = Forwarder(host, port, ioloop=self.ioloop)
        proxy.start()
        local_port = proxy.get_port()
        url = _get_proxy_url(local_port)
        worker = {"forwarder": proxy,
                  "url": url,
                  "local_port": local_port}
        return worker

    def get_proxies(self, workers):
        """Returns a dict of worker endpoints as keys, mapped to a dict
        containing the worker proxy url and local port it's mapped to.
        """
        rval = {}
        if not workers:
            workers = list(self.workers.keys())
        if workers:
            if type(workers) is str:
                workers = [workers]
        for worker in workers:
            rval[worker] = {"url": self.workers[worker]["url"],
                            "local_port": self.workers[worker]["local_port"]}
        return rval

    def __repr__(self):
        s = "ClusterProxy {name}:".format(name=_get_hostname())
        s += "\n  Scheduler: {url}".format(url=self.scheduler)
        sw = self.workers
        if sw:
            s = s+"\n  Workers:"
        for worker in sw:
            s += "\n    {worker}: {url}".format(worker=worker,
                                                url=sw[worker]["url"])
        return s

    def _repr_html_(self):
        s = "<h3>ClusterProxy {name}</h3>".format(name=_get_hostname())
        s += "<h4>Scheduler: {url}</h4>".format(url=self.scheduler)
        if len(self.workers) > 0:
            s += "<h4>Workers</h4><dl>"
            sw = self.workers
            for worker in sw:
                s += "<dt>{w}</dt><dd>{u}</dd>".format(w=worker,
                                                       u=sw[worker]["url"])
            s += "</dl>"
        return s



"""
Setup Module for LSST JupyterLab Utilities
"""
import codecs
import io
import os
import setuptools


def get_version(file, name='__version__'):
    """Get the version of the package from the given file by
    executing it and extracting the given `name`.
    """
    path = os.path.realpath(file)
    version_ns = {}
    with io.open(path, encoding="utf8") as f:
        exec(f.read(), {}, version_ns)
    return version_ns[name]


def local_read(filename):
    """Convenience function for includes"""
    full_filename = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        filename)
    return codecs.open(full_filename, 'r', 'utf-8').read()


NAME = 'jupyterlabutils'
DESCRIPTION = 'LSST utilites for JupyterLab environment'
LONG_DESCRIPTION = local_read("README.md")
VERSION = get_version('%s/_version.py' % NAME)
AUTHOR = 'Adam Thornton'
AUTHOR_EMAIL = 'athornton@lsst.org'
URL = 'https://github.com/sqre-lsst/jupyterlabutils'
LICENSE = 'MIT'

setuptools.setup(
    name=NAME,
    version=get_version("%s/_version.py" % NAME),
    long_description=LONG_DESCRIPTION,
    packages=setuptools.find_packages(),
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='lsst',
    install_requires=[
        'bokeh>=0.12.15',
        'maproxy',
        'dask',
        'distributed'
    ],
)
